/*> cylon.c <*/
/*---------------------------------------------------------------------------*/
/* This file attaches a simple interrupt handler to the standard
 * system timer interrupt. It does not replace the Kernel handler. It
 * shows how the priority field of the event structure can be used to
 * provide information only events, without the need to actually
 * handle/clear the interrupt.
 *
 * NOTE: This example relies on the LEDs being available for user
 * programs. i.e. the I/O Server option "arm_user_leds" should be
 * specified in the host.con file.
 */
/*---------------------------------------------------------------------------*/
/* standard include files */
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <syslib.h>

/* Helios-ARM specific include files */
#include <intr.h>          /* Kernel event structures */
#include <ARM/arm.h>       /* ARM specific definitions */
#include <ARM/vy86pid.h>   /* Target hardware specific definitions */

/*---------------------------------------------------------------------------*/

IntrHandler cylon ; /* provided as a global to allow access by the signal routine */

/*---------------------------------------------------------------------------*/

/* The SCR (scratch) register in the VL86C551 serial device is used as
 * the state storage area for the LED heartbeat. The following bit
 * allocations are used:
 *  b0-b1  active LED
 *  b2     direction
 *  b3-b7  count until next LED change
 */
#define hb_lpmask (0x3)     /* LED position mask */
#define hb_dir    (1 << 2)  /* direction bit */
#define hb_inc    (1 << 3)  /* lowest bit of count */
#define hb_oflow  (1 << 8)  /* first bit outside the byte */
/* Using this simple count means we only change the LED every 32 timer
 * events. On a standard 10ms timer this gives around one third of a
 * second between each change.
 *
 * Also, using a global location as our status control means that
 * multiple invocations of the cylon program will all work with the
 * same information. This means that two cylon programs will make the
 * LED move twice as quick.
 */

#pragma no_check_stack

static word cylon_interrupt(void *priv,int index)
{
 word value ;

 value = hw_SERIAL->SCR ;
 value += hb_inc ; /* increment the counter */

 if (value & hb_oflow) /* check for overflow */
  {
   byte leds ;

   value += ((value & hb_dir) ? 1 : -1) ;
   leds = (byte)(value & hb_lpmask) ; /* get pure LED number */

   if ((leds == 0) || (leds == 3))
    value ^= hb_dir ; /* change direction flag */

   leds = ~(1 << (leds + 4)) ;
   hw_PARALLEL->GPIO = leds ;
  }

 hw_SERIAL->SCR = (byte)value ;

 /* We ALWAYS return FALSE, notifying the Kernel event handler that
  * the interrupt has not been cleared. This allows the Kernel to pass
  * the event onto another handler which will continue processing it.
  */
 return(FALSE) ;
}

#pragma check_stack

/*---------------------------------------------------------------------------*/

static void terminate(int sig)
{
 printf("\nCylon terminated (signal %d)\n",sig) ;

 /* NOTE: This shows up the flaw that there is no way of cleanly
  * removing an interrupt handler automatically.
  */
 RemIntrHandler(&cylon) ;

 /* Terminate the program */
 exit(0) ;
}
/*---------------------------------------------------------------------------*/

int main(int argc,char **argv)
{
 int omode ;

 signal(SIGINT,terminate) ;

 /* NOTE: On the VY86PID we need to be in SVC mode to access the I/O space: */
 omode = ARM_GetMode() ;    /* get current processor mode */
 ARM_SwitchMode(SVCmodeG) ; /* switch to (generic) SVC mode */
 hw_SERIAL->SCR = 0 ;       /* initialise the hardware */
 ARM_SwitchMode(omode) ;    /* and return to our original mode */

 printf("Cylon demonstration (use ^C to terminate)\n") ;

 cylon.Id   = INTR_IRQ_TIMER ; /* the standard IRQ timer id */
 cylon.Pri  = 1 ; /* ensure we are executed before the system timer handler */
 cylon.Code = cylon_interrupt ; /* handler routine */
 cylon.Data = NULL ; /* no private data */
 SetIntrHandler(&cylon) ;
 /* Since the timer interrupt is already active we do not need to perform
  * any more actions. Normally the interrupt source would need to be
  * enabled, along with any other hardware setup required.
  */

 for (;;) /* loop until the user terminates this task */
  {
   printf(".") ; fflush(stdout) ;
   Delay(OneSec) ;
  }

 /* never reached */
 return(0) ;
}

/*---------------------------------------------------------------------------*/
/*> EOF cylon.c <*/
