/*> tFIQ.c <*/
/*---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
/* #include <helios.h> */
#include <kernel.h>
#include <ARM/armfiq.h>

extern void testFIQhandler(void) ; /* example FIQ routine */
extern void genFIQ(void) ;         /* simulate FIQ call routine */
extern int  testFIQhandler_size ;  /* size of the FIQ routine */

/*---------------------------------------------------------------------------*/

#define DEFAULTRETURN (0xE25EF004)  /* SUBS pc,r14_fiq,#4 */

/* Number of available FIQ ID's is defined by ARM_MAX_FIQID in armfiq.h */
typedef enum somefiqids {
 fiqid_slot0 = (0),
 fiqid_slot1 = (1),
 fiqid_slot2 = (2),
 fiqid_slot3 = (3),
 fiqid_slot4 = (4),
 fiqid_slot5 = (5),
 fiqid_slot6 = (6),
 fiqid_slot7 = (7)
} somefiqids ;

/*---------------------------------------------------------------------------*/

char *decodeFIQrcode(ARM_FIQrcodes rcode)
{
 char *cptr = "<unknown rcode>" ;
 switch (rcode)
  {
   case FIQ_OK : cptr = "FIQ_OK" ; break ;
   case FIQ_NoShare : cptr = "FIQ_NoShare" ; break ;
   case FIQ_TooSmall : cptr = "FIQ_TooSmall" ; break ;
   case FIQ_TooLarge : cptr = "FIQ_TooLarge" ; break ;
   case FIQ_Active : cptr = "FIQ_Active" ; break ;
   case FIQ_BadCode : cptr = "FIQ_BadCode" ; break ;
   case FIQ_Full : cptr = "FIQ_Full" ; break ;
  }
 return(cptr) ;
}

/*---------------------------------------------------------------------------*/

static void decodeFIQstate(void)
{
 ExecRoot *xroot = GetExecRoot() ;
 Semaphore *fsem = &(xroot->cpu.FIQ_single) ;
 int      *fflags = (int *)&(xroot->cpu.FIQ_flags) ;

 IntsOff() ;
 printf("Status: FIQ semaphore count %d : flags 0x%08X\n",(int)(fsem->Count),*fflags) ;
 IntsOn() ;
 return ;
}

/*---------------------------------------------------------------------------*/

static void decodeFIQvector(void)
{
 unsigned int *fv = (unsigned int *)0x1C ;
 printf("Status: FIQ vector contains 0x%08X\n",*fv) ;
 return ;
}

/*---------------------------------------------------------------------------*/

int main(int argc,char **argv)
{
 fncast         ansibodge ;
 word           rset[6] ; /* r8..r13 */
 word          *frset ;
 word           limit ;
 word           defcode = DEFAULTRETURN ;
 ARM_FIQrcodes  rcode ;

 /* Setup the default register values */
 rset[0] = 0x08080808 ;
 rset[1] = 0x09090909 ;
 rset[2] = 0x0A0A0A0A ;
 rset[3] = 0x0B0B0B0B ;
 rset[4] = 0x0C0C0C0C ;
 rset[5] = ARM_DefaultFIQStack() ;

 decodeFIQstate() ;
 decodeFIQvector() ;
 printf("Status: ARM_DefaultFIQStack = 0x%08X\n",(int)rset[5]) ;

 ansibodge.vfn = testFIQhandler ;
 printf("Status: testFIQhandler code at 0x%08X, size %d bytes\n",(int)(ansibodge.w),testFIQhandler_size) ;

 /* Check limits */
 rcode = ARM_AttachSFIQ(&defcode,0,NULL,rset,0) ;
 if (rcode != FIQ_TooSmall)
  {
   printf("Zero length code check failed : %s\n",decodeFIQrcode(rcode)) ;
   goto tidyup ;
  }
 printf("passed: single FIQ_TooSmall check\n") ;

 rcode = ARM_AttachSFIQ(&defcode,(4 << 10),NULL,rset,0) ;
 if (rcode != FIQ_TooLarge)
  {
   printf("Large code length check failed : %s\n",decodeFIQrcode(rcode)) ;
   goto tidyup ;
  }
 printf("passed: single FIQ_TooLarge check\n") ;

 /* Check that we can attach a single handler, and that the error and
  * timeout cases work.
  */
 printf("About to attach 1st single FIQ handler\n") ;
 rcode = ARM_AttachSFIQ(ansibodge.vp,testFIQhandler_size,NULL,rset,0) ;
 if (rcode != FIQ_OK)
  {
   printf("  Failed to attach 1st single handler : %s (%d)\n",decodeFIQrcode(rcode),rcode) ;
   goto tidyup ;
  }
 printf("  passed: 1st single FIQ handler attach\n") ;
 decodeFIQstate() ;
 decodeFIQvector() ;

 printf("About to attempt attaching another single handler\n") ;
 rcode = ARM_AttachSFIQ(&defcode,sizeof(int),NULL,rset,0) ;
 if (rcode != FIQ_Active)
  {
   printf("  Failed to return correct error when attaching single whilst single active : %s (%d)\n",decodeFIQrcode(rcode),rcode) ;
   goto tidyup ;
  }
 printf("  passed: attach single while single in use\n") ;

#if 0 /* 940418 - check multi-use attempts failing due to TimedWait problem */
 printf("About to attempt single attach with timeout (%d usecs)\n",(int)(10 * OneSec)) ;
 rcode = ARM_AttachSFIQ(&defcode,sizeof(int),NULL,rset,(10 * OneSec)) ;
 if (rcode != FIQ_Active)
  {
   printf("  Failed to return correct error when attaching single (with timeout) whilst single active : %s (%d)\n",decodeFIQrcode(rcode),rcode) ;
   goto tidyup ;
  }
 printf("  passed: attach single (with timeout) while single in use\n") ;
#endif

 ARM_ReleaseFIQ(-1) ;
 printf("1st single FIQ handler released\n") ;
 decodeFIQstate() ;
 decodeFIQvector() ;

 printf("Other single FIQ handler checks to be coded\n") ;
 /* check timeouts operate correctly */

 /* Check that we can attach multiple FIQ handlers */
 printf("About to attach the 1st multiple FIQ handler\n") ;
 rcode = ARM_AttachFIQ(fiqid_slot0,ansibodge.vp,testFIQhandler_size,NULL,&frset,&limit,0) ;
 if (rcode != FIQ_OK)
  {
   printf("  Failed to attach 1st multiple FIQ handler : %s (%d)\n",decodeFIQrcode(rcode),rcode) ;
   goto tidyup ;
  }
 printf("  passed: 1st multiple attach: register set at 0x%08X, upper-stack-limit 0x%08X\n",(int)frset,(int)limit) ; 
 decodeFIQstate() ;
 decodeFIQvector() ;
 frset[0] = 0x08080808 ;
 frset[1] = 0x09090909 ;
 frset[2] = 0x0A0A0A0A ;
 frset[3] = 0x0B0B0B0B ;
 frset[4] = 0x0C0C0C0C ;
 frset[5] = limit ; /* no good for multiple systems */
 /* We can work out how big our stack base is since we know that each
  * FIQ slot has (ARM_DefaultFIQStack - 0x1C) bytes. We also know how
  * large our code is, so we can take the delta from "limit" and get
  * an address for the first free private location for this multiple
  * FIQ slot.
  */

 printf("About to simulate a FIQ event\n") ;
*((unsigned int *)0x1000) = 0x00001000 ; /* LSA trigger */
 genFIQ() ;
 printf("  passed: simulated FIQ\n") ;

 printf("About to attempt attaching single handler\n") ;
 rcode = ARM_AttachSFIQ(&defcode,sizeof(int),NULL,rset,0) ;
 if (rcode != FIQ_NoShare)
  {
   printf("  Failed to return correct error when attaching single whilst multiple active : %s (%d)\n",decodeFIQrcode(rcode),rcode) ;
   goto tidyup ;
  }
 printf("  passed: attempt to attach single while multiple in use\n") ;
 decodeFIQstate() ;

 printf("About to attach the 2nd multiple FIQ handler\n") ;
 rcode = ARM_AttachFIQ(fiqid_slot1,ansibodge.vp,testFIQhandler_size,NULL,&frset,&limit,0) ;
 if (rcode != FIQ_OK)
  {
   printf("  Failed to attach 2nd multiple FIQ handler : %s (%d)\n",decodeFIQrcode(rcode),rcode) ;
   goto tidyup ;
  }
 printf("  passed: 2nd multiple attach: register set at 0x%08X, stack-limit 0x%08X\n",(int)frset,(int)limit) ;
 decodeFIQstate() ;

 printf("More multiple FIQ handler checks to be coded\n") ;
 /* .. allocate another multiple FIQ slot .. */

 printf("About to release multiple FIQ ID 0x%08X\n",fiqid_slot0) ;
 ARM_ReleaseFIQ(fiqid_slot0) ;
 printf("About to release multiple FIQ ID 0x%08X\n",fiqid_slot1) ;
 ARM_ReleaseFIQ(fiqid_slot1) ;
 decodeFIQstate() ;
 decodeFIQvector() ;

 printf("About to attempt attaching single handler\n") ;
 rcode = ARM_AttachSFIQ(ansibodge.vp,testFIQhandler_size,NULL,rset,0) ;
 if (rcode != FIQ_OK)
  {
   printf("  Failed to attach single FIQ handler : %s (%d)\n",decodeFIQrcode(rcode),rcode) ;
   goto tidyup ;
  }
 printf("  passed: attempt to attach single\n") ;
 decodeFIQstate() ;
 decodeFIQvector() ;

 printf("About to release single FIQ handler\n") ;
 ARM_ReleaseFIQ(-1) ;
 decodeFIQstate() ;
 decodeFIQvector() ;

 printf("Tests completed OK\n") ;
 return(0) ;

tidyup:

 printf("ERROR: About to attempt release of all the possible FIQ sources\n") ;

 ARM_ReleaseFIQ(-1) ;
 ARM_ReleaseFIQ(fiqid_slot0) ;
 ARM_ReleaseFIQ(fiqid_slot1) ;
 ARM_ReleaseFIQ(fiqid_slot2) ;
 ARM_ReleaseFIQ(fiqid_slot3) ;
 ARM_ReleaseFIQ(fiqid_slot4) ;
 ARM_ReleaseFIQ(fiqid_slot5) ;
 ARM_ReleaseFIQ(fiqid_slot6) ;
 ARM_ReleaseFIQ(fiqid_slot7) ;

 exit(1) ;
}

/*---------------------------------------------------------------------------*/
/*> EOF tFIQ.c <*/
